from threading import Thread
from logger_tt import setup_logging, getLogger

setup_logging()
logger = getLogger(__name__)


def faulty_func(a=1, b=0):
    c = a/b
    return c


if __name__ == '__main__':
    # create an error in another thread and try to log it
    t = Thread(target=faulty_func)
    t.start()
    t.join()

    # can we continue the logging ?
    logger.debug('The thread has finished')



import re
import sys

import pytest
from subprocess import run, PIPE
from pathlib import Path

from logger_tt import setup_logging

__author__ = "Duc Tin"

log = Path.cwd() / 'logs/log.txt'


def test_multiprocessing_normal():
    cmd = [sys.executable, "multiprocessing_normal.py", "3"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    assert 'Parent process is ready to spawn child' in data
    assert 'child process 0' in data
    assert 'child process 1' in data
    assert 'child process 2' in data


@pytest.mark.parametrize('value', [-1, 2, 'forker', 'spawm'])
def test_multiprocessing_error(value):
    with pytest.raises(ValueError) as e:
        setup_logging(use_multiprocessing=value)

    assert f'Expected a bool or a multiprocessing start_method name, but got: {value}' in str(e)


def test_multiprocessing_pool():
    cmd = [sys.executable, "multiprocessing_pool.py", "10"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    assert 'Parent process pool is ready to spawn child' in data
    assert 'child process 0' in data
    assert 'child process 1' in data
    assert 'child process 2' in data
    assert 'child process 3' in data
    assert 'child process 4' in data
    assert 'child process 5' in data
    assert 'child process 6' in data
    assert 'child process 7' in data
    assert 'child process 8' in data
    assert 'child process 9' in data


def test_multiprocessing_threading():
    """Test a default logger"""
    cmd = [sys.executable, "multiprocessing_threading.py", "10"]
    result = run(cmd, stdout=PIPE, universal_newlines=True)
    assert 'Parent process is ready to spawn child' in result.stdout
    expect = re.findall(r'Process-\d+ Thread-\d+.*? thread running from process', result.stdout)
    assert len(expect) == 10

    data = log.read_text(encoding='utf8')
    assert 'Parent process is ready to spawn child' in data
    expect = re.findall(r'Process-\d+ Thread-\d+.*? thread running from process', data)
    assert len(expect) == 10



from logging import getLogger
from logger_tt import setup_logging

__author__ = "Duc Tin"

setup_logging(capture_print=True)
logger = getLogger(__name__)


class Dummy:
    def __str__(self):
          return f"Ahahaha\n ehehe \n lalala"


class Tummy:
    def __init__(self):
        self.dummy = Dummy()


def aloha(a):
    b = Dummy()
    c = b / a.dummy


if __name__ == '__main__':
    a = Tummy()
    aloha(a)



import time
import sys
from random import randint

from multiprocessing import Process
from threading import Thread
from logger_tt import setup_logging, logger


__author__ = "Duc Tin"
setup_logging(use_multiprocessing=True)


def threader(arg):
    logger.info(f'thread running from process {arg}')
    time.sleep(randint(1, 10) / 10)


def worker(arg):
    logger.info(f'child process {arg}: started')
    th = Thread(target=threader, args=(arg,))
    th.start()
    th.join()
    logger.info(f'child process {arg}: stopped')


if __name__ == '__main__':
    if len(sys.argv) > 1:
        proc_no = int(sys.argv[1])
    else:
        proc_no = 7

    all_processes = []
    logger.info('Parent process is ready to spawn child')
    for i in range(proc_no):
        p = Process(target=worker, args=(i,))
        all_processes.append(p)
        p.daemon = True
        p.start()

    for p in all_processes:
        p.join()

    print('__finished__')



from pathlib import Path

__author__ = "Duc Tin"


log = Path.cwd() / 'logs/log.txt'


def remove_log():
    log_folder = log.parent
    if log_folder.exists():
        for file in log_folder.iterdir():
            file.unlink()
        # else:
        #     log_folder.unlink()


def pytest_runtest_setup(item):
    """call before execute a test of item"""
    remove_log()


def pytest_runtest_teardown(item, nextitem):
    """call after executed a test item"""
    import logging
    logging.shutdown()
    remove_log()


def pytest_configure(config):
    """
    Allows plugins and conftest files to perform initial configuration.
    This hook is called for every plugin and initial conftest
    file after command line options have been parsed.
    """
    pass


def pytest_sessionstart(session):
    """
    Called after the Session object has been created and
    before performing collection and entering the run test loop.
    """
    pass


def pytest_sessionfinish(session, exitstatus):
    """
    Called after whole test run finished, right before
    returning the exit status to the system.
    """
    pass


def pytest_unconfigure(config):
    """
    called before test process is exited.
    """
    pass



from logging import getLogger
from logger_tt import setup_logging

__author__ = "Duc Tin"

setup_logging()
logger = getLogger(__name__)

try:
    a = 1
    b = 0
    c = a / b
except Exception as e:
    logger.exception('some error')



import socket
import sys
import logging
import atexit
import platform
from logging import handlers
from multiprocessing import Queue as mpQueue, current_process
from queue import Queue as thQueue
from threading import Thread, main_thread

import pickle
import socketserver
import struct
import select

from .capture import PrintCapture

__author__ = "Duc Tin"
root_logger = logging.getLogger()


class LogConfig:
    def __init__(self):
        self.qclass = None
        self.queues = []
        self.q_listeners = []
        self.converted_handlers = set()

        # tcp port for socket handler
        self.host = 'localhost'
        self.port = handlers.DEFAULT_TCP_LOGGING_PORT
        self.tcp_server = None

        # other settings
        self.full_context = False
        self.__capture_print = False
        self.strict = False
        self.guess_level = False
        self.suppress_level_below = logging.WARNING

        self.original_stdout = sys.stdout

        # suppress logger list for usage in filter
        self.suppress_list = set()

        # initialized counter
        self.__initialized = 0

    @property
    def initialized(self):
        return self.__initialized

    def from_dict(self, odict: dict):
        # store basic settings
        for key in ['full_context', 'strict', 'guess_level']:
            setattr(self, key, odict[key])

        # set capture_print
        self.capture_print = odict['capture_print']

        # suppress other logger
        level = odict.get('suppress_level_below', 'WARNING')
        if type(level) is str:
            if level.upper() not in ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']:
                raise ValueError(f'"level" string is incorrect: {level}')
            level = getattr(logging, level.upper())
        self.suppress_level_below = level
        self.suppress_loggers(odict.get("suppress"))

        # set logging mode accordingly
        self.set_mode(odict['use_multiprocessing'])

        self.__initialized += 1

    def set_mode(self, use_multiprocessing):
        """Select logging method according to platform and multiprocessing"""
        os_name = platform.system()
        if use_multiprocessing not in [True, False, 'fork', 'spawn', 'forkserver']:
            raise ValueError(f'Expected a bool or a multiprocessing start_method name, but got: {use_multiprocessing}')

        if not use_multiprocessing:
            # for normal usage, thread queue is more than enough
            self.qclass = thQueue
            self.replace_with_queue_handler()
        else:
            # multiprocessing
            if os_name == 'Linux' and use_multiprocessing == 'fork':
                # because of copy on write while forking, multiprocessing queue can be used
                self.qclass = mpQueue
                self.replace_with_queue_handler()
            else:
                # __main__ is imported from crash for each child process
                # so we must use socket to communicate between processes
                self.replace_with_socket_handler()

    def replace_with_queue_handler(self):
        """ setup a central queue handler and start a listener thread """
        all_loggers = [root_logger] + [logging.getLogger(name) for name in root_logger.manager.loggerDict]

        for logger in all_loggers:
            if not logger.handlers:
                continue

            # backup current handlers then clear it
            all_handlers = logger.handlers
            logger.handlers = []

            # add queue handler
            queue = self.qclass()
            q_handler = handlers.QueueHandler(queue)
            logger.addHandler(q_handler)

            ql = handlers.QueueListener(queue, *all_handlers, respect_handler_level=True)
            self.q_listeners.append(ql)

            # start listening
            atexit.register(ql.stop)
            ql.start()

        root_logger.debug('Logging queue listener started!')

    def replace_with_socket_handler(self):
        """ setup a central socket handler and start a listener server """
        logger = logging.getLogger()

        # backup current handlers
        all_handlers = logger.handlers

        # clear all handlers
        logger.handlers = []

        # add socket handler
        socket_handler = logging.handlers.SocketHandler(self.host, self.port)
        atexit.register(socket_handler.close)
        logger.addHandler(socket_handler)

        # initiate server
        if current_process().name == 'MainProcess':
            self.tcp_server = LogRecordSocketReceiver(self.host, self.port, all_handlers)
            serving = Thread(target=self.tcp_server.serve_until_stopped)
            serving.start()
            root_logger.debug('Logging server started!')

    def suppress_loggers(self, loggers):
        if not loggers:
            return

        for name in loggers:
            logger = logging.getLogger(name)
            logger.level = self.suppress_level_below

        self.suppress_list.update(loggers)

    @property
    def capture_print(self):
        return self.__capture_print

    @capture_print.setter
    def capture_print(self, val):
        if val:
            sys.stdout = PrintCapture(sys.stdout, strict=self.strict, guess_level=self.guess_level)
        else:
            sys.stdout = self.original_stdout

    def __enter__(self):
        """This is to simplify pytest test case"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """This is to simplify pytest test case"""
        # self.queue.join()
        while self.q_listeners:
            ql = self.q_listeners.pop()
            ql.stop()
            atexit.unregister(ql.stop)


class LogRecordStreamHandler(socketserver.StreamRequestHandler):
    """Handler for a streaming logging request.

    This basically logs the record using whatever logging policy is
    configured locally.
    """
    # log record handlers
    handlers = []

    def receive_meta(self):
        """Get the byte length of incoming log record"""
        chunk = bytearray()

        while len(chunk) < 4:
            try:
                b = self.connection.recv(1)
                if not b:
                    # client disconnected
                    return ''
                else:
                    chunk.extend(b)
            except socket.timeout:
                if main_thread().is_alive():
                    continue
                else:
                    # Main thread exited but
                    # the client of from main thread hasn't disconnected yet.
                    # We break the loop to avoid dead-lock
                    return b''
            except ConnectionResetError as e:
                # connection was forcibly closed by client
                return b''

        return chunk

    def handle(self):
        """
        Handle multiple requests - each expected to be a 4-byte length,
        followed by the LogRecord in pickle format. Logs the record
        according to whatever policy is configured locally.
        """
        self.connection.settimeout(5)
        while True:
            chunk = self.receive_meta()
            if not chunk:
                break

            record_len = struct.unpack('>L', chunk)[0]
            chunk = self.connection.recv(record_len)
            while len(chunk) < record_len:
                chunk = chunk + self.connection.recv(record_len - len(chunk))

            # unpickle data
            obj = pickle.loads(chunk)
            record = logging.makeLogRecord(obj)

            # handle
            self.handle_log_record(record)

    def handle_log_record(self, record):
        """Handle a record.
            This just loops through the handlers offering them the record
            to handle.
        """
        for handler in self.handlers:
            process = record.levelno >= handler.level
            if process:
                handler.handle(record)


class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    """
    Simple TCP socket-based logging receiver suitable for testing.
    """

    allow_reuse_address = True
    daemon_threads = False  # Set this to True to immediate exit when main-thread exits.

    # There is a chance that it terminates some log records that are
    # being processed

    def __init__(self, host, port, log_record_handlers):
        LogRecordStreamHandler.handlers = log_record_handlers
        super().__init__((host, port), LogRecordStreamHandler)
        self.abort = False
        self.timeout = 1

    def serve_until_stopped(self):
        while main_thread().is_alive():
            rd, wr, ex = select.select([self.socket.fileno()],
                                       [], [],
                                       self.timeout)
            if rd:
                self.handle_request()


class DefaultFormatter(logging.Formatter):
    """Based on the format string of any handler in root, we make new formatters for
        the default logger - logger_tt. This is a way to inject information but
        using the same handlers as regular logger.
    """
    default_formats = dict(normal=["%(name)s", "%(filename)s"],
                           thread=["%(message)s", "%(threadName)s %(message)s"],
                           multiprocess=["%(message)s", "%(processName)s %(message)s"],
                           both=["%(message)s", "%(processName)s %(threadName)s %(message)s"])

    def __init__(self, fmt: str = '', datefmt: str = '', style: str = ''):
        super(DefaultFormatter, self).__init__(fmt=fmt, datefmt=datefmt)

        self._logger_tt_formatters = {}
        for case, fmt in self._standardize(fmt).items():
            self._logger_tt_formatters[case] = logging.Formatter(fmt=fmt, datefmt=datefmt, style=style)

    def _standardize(self, fmt):
        formatters = {'normal': fmt.replace(self.default_formats['normal'][0], self.default_formats['normal'][1])}

        # concurrent format
        concurrent_fmt = formatters['normal'].replace('%(threadName)s', '').replace('%(processName)s', '')
        for _type, replacement in self.default_formats.items():
            if _type == 'normal':
                continue

            old, new = replacement
            formatters[_type] = concurrent_fmt.replace(old, new)

        return formatters

    def format(self, record):
        if record.name == 'logger_tt':
            if record.processName == 'MainProcess' and record.threadName == 'MainThread':
                return self._logger_tt_formatters['normal'].format(record)
            elif record.processName == 'MainProcess' and record.threadName != 'MainThread':
                return self._logger_tt_formatters['thread'].format(record)
            elif record.processName != 'MainProcess' and record.threadName == 'MainThread':
                return self._logger_tt_formatters['multiprocess'].format(record)
            else:
                return self._logger_tt_formatters['both'].format(record)

        return super(DefaultFormatter, self).format(record)



from logger_tt import setup_logging
from logging import getLogger


__author__ = "Duc Tin"
setup_logging()
logger = getLogger(__name__)


class Dummy:
    def __init__(self):
        self.value = 3
        self.divisor = 0


if __name__ == '__main__':
    a = Dummy()
    res = a.value / a.divisor
    print(res)



import sys
from subprocess import run
from pathlib import Path


__author__ = "Duc Tin"

log = Path.cwd() / 'logs/log.txt'


def test_log_uncaught_thread_exception():
    cmd = [sys.executable, "exception_thread.py"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    assert 'Uncaught exception' in data
    assert 'The thread has finished' in data



import time
from logging import getLogger
from logger_tt import setup_logging

__author__ = "Duc Tin"

setup_logging(full_context=True)
logger = getLogger(__name__)


class MyClass:
    def __init__(self):
        self.a = "aloha"

    def __del__(self):
        time.sleep(1)
        logger.info(f"delete objects")


if __name__ == '__main__':
    c = MyClass()
    d = 5 / 0



__author__ = "Duc Tin"

from logging import getLogger

logger = getLogger('exchangelib')


def fox_run():
    logger.debug('Exchangelib module started')
    logger.info('A quick brown fox jumps over a flower fence')
    logger.warning('A fence is too high')
    logger.error("A fox's rear legs stuck in the fence")
    logger.critical("A fox is bleeding")



import logging
import re
import sys
from logging import getLogger
from pathlib import Path

import pytest
from logger_tt import setup_logging, logging_disabled, logger as my_logger


__author__ = "Duc Tin"

logger = getLogger(__name__)
log = Path.cwd() / 'logs/log.txt'


def test_basic_function(capsys):
    with setup_logging():
        logger.debug('my debug')
        logger.info('my info')
        logger.warning('my warning')
        logger.error('my error')
        logger.critical('the market crashed')

    # check stdout
    captured = capsys.readouterr()
    stdout_data = captured.out
    assert 'my debug' not in stdout_data
    assert 'my info' in stdout_data
    assert 'my warning' in stdout_data
    assert 'my error' in stdout_data
    assert 'the market crashed' in stdout_data

    # check log.txt
    log_data = log.read_text()
    assert 'my debug' in log_data
    assert 'my info' in log_data
    assert 'my warning' in log_data
    assert 'my error' in log_data
    assert 'the market crashed' in log_data


def test_capture_print_not_strict(capsys):
    with setup_logging(capture_print=True):
        print('This is my print')
        sys.stdout.write('\n')
        sys.stdout.write('This should be printed normally')

    stdout_data = capsys.readouterr().out
    log_data = log.read_text()

    assert re.search('.*INFO.*This is my print', stdout_data)
    assert re.search('.*INFO.*This is my print', log_data)

    assert 'This should be printed normally' in stdout_data
    assert not re.search('.*INFO.*This should be printed normally', stdout_data)
    assert 'This should be printed normally' not in log_data


def test_capture_print_strict(capsys):
    with setup_logging(capture_print=True, strict=True):
        print('This is my print')
        sys.stdout.write('This should be printed normally too')

    stdout_data = capsys.readouterr().out
    log_data = log.read_text()

    assert re.search('.*INFO.*This is my print', stdout_data)
    assert re.search('.*INFO.*This is my print', log_data)

    assert re.search('.*INFO.*This should be printed normally too', stdout_data)
    assert re.search('.*INFO.*This should be printed normally too', log_data)


@pytest.mark.parametrize("msg", [('info', 'abc Info: It will rain this afternoon'),
                                 ('warning', 'def Warning: the price is down'),
                                 ('error', 'ghi Error: username incorrect'),
                                 ('critical', 'jkl Critical: system is overheating'),
                                 ('debug', 'mno DEBUG: ha ha ha')])
def test_guess_message_level(msg):
    with setup_logging(capture_print=True, guess_level=True):
        level, msg = msg
        print(msg)

    log_data = log.read_text().splitlines()[-1].lower()
    assert log_data.count(level) == 2


@pytest.mark.parametrize("level", [logging.WARNING, logging.ERROR])
def test_suppress_logger(capsys, level):
    # suppess by config file
    with setup_logging(suppress_level_below=level):
        from tests.exchangelib_logger import fox_run

        fox_run()

    stdout_data = capsys.readouterr().out
    assert 'DEBUG' not in stdout_data
    assert 'INFO' not in stdout_data

    if level == logging.WARNING:
        assert 'WARNING' in stdout_data
    elif level == logging.ERROR:
        assert 'WARNING' not in stdout_data

    assert 'ERROR' in stdout_data
    assert 'CRITICAL' in stdout_data


def test_suppress_logger2(capsys):
    # suppress by the code
    with setup_logging(suppress=['suppressed_logger']):
        logger = getLogger('suppressed_logger')
        logger.info('Ha ha, this should not be logged')

    stdout_data = capsys.readouterr().out
    assert 'INFO' not in stdout_data


def test_logging_disabled(capsys):
    with setup_logging():
        logger.info('Secret process starts')
        with logging_disabled():
            logger.debug('debug')
            logger.info('info')
            logger.warning('warning')
            logger.error('error')
            logger.critical('critical')
        logger.info('Secret process finished')

    stdout_data = capsys.readouterr().out
    assert 'debug' not in stdout_data
    assert 'info' not in stdout_data
    assert 'warning' not in stdout_data
    assert 'error' not in stdout_data
    assert 'critical' not in stdout_data

    log_data = log.read_text()
    assert 'debug' not in log_data
    assert 'info' not in log_data
    assert 'warning' not in log_data
    assert 'error' not in log_data
    assert 'critical' not in log_data


def test_default_logger(capsys):
    with setup_logging():
        my_logger.debug('debug')
        my_logger.info('info')
        my_logger.warning('warning')
        my_logger.error('error')
        my_logger.critical('critical')

    stdout_data = capsys.readouterr().out
    assert 'debug' not in stdout_data
    assert 'info' in stdout_data
    assert 'warning' in stdout_data
    assert 'error' in stdout_data
    assert 'critical' in stdout_data

    log_data = log.read_text()
    assert re.search(r'test_simple:\d+.+debug', log_data)
    assert re.search(r'test_simple:\d+.+info', log_data)
    assert re.search(r'test_simple:\d+.+warning', log_data)
    assert re.search(r'test_simple:\d+.+error', log_data)
    assert re.search(r'test_simple:\d+.+critical', log_data)


def test_default_logger_submodule():
    with setup_logging():
        from tests.sub_module import run

        run()

    log_data = log.read_text()
    assert len(re.findall(r"tests.sub_module:\d+ INFO", log_data)) == 2


def test_default_logger_suppress():
    with setup_logging() as log_config:
        from tests.sub_module import run

        log_config.suppress_loggers(['tests.sub_module'])
        run()

    log_data = log.read_text()
    assert len(re.findall(r"tests.sub_module:\d+ INFO", log_data)) == 0


def test_double_setup_logging():
    # there should be a warning if setup_logging() is called multiple times
    with setup_logging() as log_config:
        with setup_logging() as log_config:
            pass

    log_data = log.read_text()
    assert "WARNING" in log_data



import re
import inspect
import logging


__author__ = "Duc Tin"


print_logger = logging.getLogger('PrintCapture')
print_level = {'debug': print_logger.debug,
               'info': print_logger.info,
               'warning': print_logger.warning,
               'warn': print_logger.warning,
               'critical': print_logger.critical,
               'error': print_logger.error,
               'err': print_logger.error,
               }


def log_level(message):
    """ guess level of msg and log it """
    msg = message.lower()[:100]     # search within the first few character only
    for keyword in print_level:
        pattern = fr'\b{keyword}\b'
        if re.search(pattern, msg):
            level = keyword
            break
    else:
        level = 'info'

    print_level[level](message)


def is_print_called():
    for frame in inspect.stack():
        context = frame.code_context
        if not context:
            continue

        context = context[0].strip()
        if re.match(r'print\(.*\)?', context):
            return True
    return False


class PrintCapture(object):
    """Man in the middle of the stream"""
    def __init__(self, original, strict: bool, guess_level=False):
        self.terminal = original        # original stream
        self.strict = strict            # True: log full stream, False: log print() message only
        self.guess_level = guess_level
        self.log = log_level if guess_level else print_logger.info

    def write(self, message):
        message = message.strip()
        if not message:
            return

        if not self.strict and is_print_called():
            # capture print() function only
            self.log(message)
        elif self.strict:
            # capture anything that use sys.stdout.write()
            self.log(message)
        else:
            # leave sys.stdout.write() alone
            self.terminal.write(message)

    def flush(self):
        pass



import re
import sys
from pathlib import Path
from subprocess import run, PIPE

import pytest

from logger_tt.inspector import get_recur_attr, get_repr, is_full_statement, get_full_statement, MEM_PATTERN


__author__ = "Duc Tin"
log = Path.cwd() / 'logs/log.txt'


def test_get_recur_attr():
    class A:
        def __init__(self):
            self.other = None

        @property
        def x(self):
            # error while getting attribute
            return 1 / 0

    a = A()
    b = A()
    a.other = b
    b.other = a

    assert get_recur_attr(a, 'other') == b
    assert get_recur_attr(a, 'other.other') == a
    assert get_recur_attr(a, 'other.other.other') == b
    assert get_recur_attr(a, 'other.other.me') == "!!! Not Exists"
    assert "!!! Attribute Error: " in get_recur_attr(a, 'other.x')


def test_get_repr():
    class A:
        def __str__(self):
            return "A 123"

    class B:
        def __repr__(self):
            return 'B(arg=arg)'

    class C:
        def __repr__(self):
            # error while getting representation
            return "123" + 3

    class D:
        def __str__(self):
            # return __repr__ in case of __str__ error
            return "123" + 3

    assert get_repr(A()) == "A 123"
    assert get_repr(B()) == "B(arg=arg)"
    assert "!!! repr error:" in get_repr(C())
    assert MEM_PATTERN.match(get_repr(D()))


def test_is_full_statement():
    pycode = "print(f'{self.value} '"
    assert not is_full_statement(pycode)

    assert is_full_statement(pycode, "      'f{another} line'", "      )")

    pycode = "print(3+4)"
    assert is_full_statement(pycode)


def test_1_scope():
    cmd = [sys.executable, "exception_main.py"]
    run(cmd)

    data = log.read_text()
    assert "a.value = 3" in data
    assert "a.divisor = 0" in data


def test_nested_1():
    cmd = [sys.executable, "exception_nested_1.py"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    assert "var_in = Dummy(my dummy class)" in data
    assert "arg = (456, 789)" in data
    assert "kwargs = {'my_kw': 'hello', 'another_kw': 'world'}" in data
    assert "my_local_var = 345" in data


def test_nested_2():
    cmd = [sys.executable, "exception_nested_2.py"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    assert "self.value = 3" in data
    assert "self.non_exist = '!!! Not Exists'" in data
    assert "self.base.name = 'Nested dot'" in data


def test_full_context_level1():
    cmd = [sys.executable, "exception_full_context.py", "1"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    assert "__name__ = '__main__'" not in data
    assert "Base = <class '__main__.Base'>" not in data
    assert "arg = (456, 789)" not in data
    assert "arg = 345" in data


def test_full_context_level2():
    cmd = [sys.executable, "exception_full_context.py", "2"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    assert "__name__ = '__main__'" not in data
    assert "Base = <class '__main__.Base'>" not in data
    assert "arg = (456, 789)" in data
    assert "arg = 345" in data


def test_full_context_level3():
    cmd = [sys.executable, "exception_full_context.py", "3"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    assert "__name__ = '__main__'" in data
    assert "Base = <class '__main__.Base'>" in data
    assert "arg = (456, 789)" in data
    assert "arg = 345" in data


def test_log_exception():
    cmd = [sys.executable, "exception_log.py"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    assert "-> a = 1" in data
    assert "-> b = 0" in data


@pytest.mark.skipif(sys.version_info >= (3, 7), reason="'tokenize' module changed behavior")
def test_multiline():
    cmd = [sys.executable, "exception_multiline.py"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    assert "self.value = 3" in data
    assert "self.non_exist = '!!! Not Exists'" in data
    assert "self.base.name = 'Nested dot'" in data


def test_multiline2():
    cmd = [sys.executable, "exception_multiline2.py"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    res = re.findall(r'( *ehehe)', data)
    indent_1 = res[0].count(' ')
    indent_2 = res[1].count(' ')

    assert indent_1 == 13
    assert indent_2 == 27


@pytest.mark.skip(reason='No example of deadlock')
def test_deadlock():
    cmd = [sys.executable, "exception_deadlock.py"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    print(data)
    # fix me
    assert 1


def test_del_logging():
    cmd = [sys.executable, "exception_del_logging.py"]
    output = run(cmd, stderr=PIPE, universal_newlines=True)
    assert "Logging error" not in output.stderr


def test_logging_class():
    # pre-existing logger should also have a new exception handler
    cmd = [sys.executable, "sub_module.py"]
    output = run(cmd, stdout=PIPE, universal_newlines=True)
    data = log.read_text()
    assert "a = 3" in data
    assert "b = 0" in data
    assert "a = 3" in output.stdout
    assert "b = 0" in output.stdout


def test_get_full_statement():
    lines = get_full_statement('full_statement.py', 5)
    assert lines == ['print(x,\n', '          y(1,\n', '            z),\n', '          2)']

    lines = get_full_statement('full_statement.py', 6)
    assert lines == ['y(1,\n', '  z),']

    lines = get_full_statement('full_statement.py', 7)
    assert lines == ['y(1,\n', '  z),']

    lines = get_full_statement('full_statement.py', 17)
    assert lines == ['txt = {a: 3,\n', '       b: 4,\n', '       c:5}']

    lines = get_full_statement('full_statement.py', 23)
    assert lines == ['my_list = [a,\n', '           b,\n', '           c]']



from logger_tt import setup_logging
from logging import getLogger


__author__ = "Duc Tin"
setup_logging()
logger = getLogger(__name__)


class Base:
    def __init__(self):
        self.name = 'Nested dot'


class Dummy:
    def __init__(self):
        self.value = 3
        self.divisor = 0
        self.base = Base()

    def details(self, arg):
        print(f'{self.value} {self.non_exist} {self.base.name}')

    def __str__(self):
        return 'Dummy(my dummy class)'


def my_function(var_in, *arg, **kwargs):
    my_local_var = 345
    var_in.details(my_local_var)


if __name__ == '__main__':
    a = Dummy()
    my_function(a, 456, 789, my_kw='hello', another_kw='world')



from logging import getLogger

__author__ = "Duc Tin"
logger = getLogger(__name__)


def error_func():
    # create an exception
    a = 3
    b = 0
    try:
        res = a / b
        return res
    except Exception as e:
        logger.exception(e)



__author__ = "Duc Tin"

import re
import logging
import tokenize
import linecache
from io import StringIO
from contextlib import contextmanager
from traceback import StackSummary, walk_tb

ID_PATTERN = re.compile(r'([a-zA-Z_][a-zA-Z_0-9.]*)')
MEM_PATTERN = re.compile(r'<.*object at 0x[0-9A-Fa-f]+>')


def get_recur_attr(obj, attr: str):
    """
    Follow the dot `.` in attribute string `attr` to the final object

    :param obj: any object
    :param attr: string of attribute access
    :return: the final desire object or '!!! Not Exists'

    example: if we need to access `c` object as in `a.b.c`,
    then `obj=a, attr='b.c'`
    If `b` or `c` doesn't exists, then `'!!! Not Exists'` is returned
    """
    this_level, *next_levels = attr.split('.', maxsplit=1)
    if not next_levels:
        try:
            return getattr(obj, this_level)
        except AttributeError:
            return "!!! Not Exists"
        except Exception as e:
            return f"!!! Attribute Error: {e}"
    else:
        try:
            obj = getattr(obj, this_level)
        except AttributeError:
            return "!!! Not Exists"
        return get_recur_attr(obj, next_levels[0])


def get_repr(obj, multiline_indent: int = 0) -> str:
    """
    Pick a more useful representation of object `obj` between __str__ and __repr__
     if it available.

    :param obj: any of object
    :param multiline_indent: indent level for the second line onward
    :return: string of object representation
    """
    try:
        _repr_ = repr(obj)
        if not MEM_PATTERN.match(_repr_):
            return _repr_
    except Exception as e:
        _repr_ = f"!!! repr error: {e}"

    try:
        _str_ = str(obj).replace('\r\n', '\n')
        if '\n' in _str_:
            indent = ' ' * multiline_indent
            _str_ = _str_.replace('\n', '\n' + indent)

        return _str_

    except Exception as e:
        return _repr_


def is_half_ended(line: str) -> bool:
    """
    Check if we a given a last line of a multiline statement with brackets
    """
    if line.count(')') > line.count('('):
        return True
    if line.count('}') > line.count('{'):
        return True
    if line.count(']') > line.count('['):
        return True


def is_full_statement(*lines: str) -> bool:
    """
    Check if a set of lines makes up a full python statement

    :param lines: list of line of python code
    :return: True if line in `lines` made up a full python statement
    """
    try:
        stream = StringIO()
        stream.write('\n'.join(lines))
        stream.seek(0)
        for t in tokenize.generate_tokens(stream.readline):
            pass
    except tokenize.TokenError as e:
        if 'EOF in multi-line' in str(e):
            return False
        else:
            return True
    else:
        return True


def get_statement_up(filename: str, lineno: int) -> list:
    """The python interpreter return the last line of exception lines
        This behavior is of python 3.6, 3.7
        We have to go up to grab the full statement

        Grab maximum 10 lines
    """
    startno = lineno
    lines = []

    while lineno >= max(0, startno - 9):
        line = linecache.getline(filename, lineno)
        lines.insert(0, line)

        temp = ''.join(lines)
        if not is_half_ended(temp):
            break

        lineno -= 1

    return lines


def get_statement_down(filename: str, lineno: int) -> list:
    """The python interpreter return the first line of exception lines
        This behavior is of python 3.9
        We go down normally to grab the full statement

        Grab maximum 10 lines
    """
    lines = []
    startno = lineno
    while lineno <= startno + 9:
        line = linecache.getline(filename, lineno)
        if not line:
            break

        lines.append(line)
        if is_full_statement(*lines):
            break

        lineno += 1

    return lines


def get_full_statement(filename, lineno: int) -> list:
    """
    Get all lines of python file `filename` that makes up a full statement starting from `lineno`

    :param filename: path to python source code file
    :param lineno: line number at with the statement started
    :return: list of maximum 5 lines that made up a python statement with indent striped,
    """

    line = linecache.getline(filename, lineno)
    if is_half_ended(line):
        lines = get_statement_up(filename, lineno)
    else:
        lines = get_statement_down(filename, lineno)

    lines[-1] = lines[-1].strip('\n')  # remove newline in last line
    indent = re.search(r'^(\s+)', lines[0])
    if indent:
        indent = indent.group(1)
        return [x.replace(indent, '', 1) for x in lines]
    else:
        return lines


@contextmanager
def logging_disabled():
    logging.disable(logging.CRITICAL)
    yield
    logging.disable(logging.NOTSET)


def get_traceback_depth(trace_back) -> int:
    count = 0
    tb = trace_back
    while tb is not None:
        tb = tb.tb_next
        count += 1

    return count


def analyze_frame(trace_back, full_context: int = False) -> str:
    """
    Read out variables' content surrounding the error line of code

    :param trace_back: A traceback object when exception occur
    :param full_context: Also export local variables that is not in the error line
            full_context == 0: export only variables that appear on the error line
            full_context == 1: export variables within the error function's scope
            full_context >= 2: export variables along the function's call stack up to `full_context` level
    :return: string of analyzed frame
    """
    result = []
    full_context = max(0, int(full_context))
    # todo: add color
    bullet_1 = '|->'
    bullet_2 = '=>'
    multi_line_indent1 = 8 + len(bullet_1)
    multi_line_indent2 = 8 + len(bullet_2)
    stack_depth = get_traceback_depth(trace_back)

    with logging_disabled():
        for idx, obj in enumerate(walk_tb(trace_back)):
            frame, _ = obj

            global_var = frame.f_globals
            local_var = frame.f_locals

            summary = StackSummary.extract([obj], capture_locals=True)[0]
            line = summary.line
            if not is_full_statement(summary.line):
                line = get_full_statement(summary.filename, summary.lineno)
                line = '    '.join(line)

            txt = [f'  File "{summary.filename}", line {summary.lineno}, in {summary.name}',
                   f'    {line}']

            # todo: dump all level to different file?
            parse_level = max(full_context - 1, 0)
            if idx + 1 < (stack_depth - parse_level):
                # don't parse variables for top levels
                txt.append('')
                result.append('\n'.join(txt))
                continue

            # get value of variables on the error line
            identifiers = ID_PATTERN.findall(line)
            seen = set()
            outer = "(outer) " if idx else ""  # ground level variables are not outer for sure
            for i in identifiers:
                if i in seen or i.endswith('.'):
                    continue

                seen.add(i)
                spaces = multi_line_indent1 + len(i)
                if i in local_var:
                    value = get_repr(local_var[i], spaces)
                    txt.append(f'     {bullet_1} {i} = {value}')
                elif i in global_var:
                    spaces += len(outer)
                    value = get_repr(global_var[i], spaces)
                    txt.append(f'     {bullet_1} {outer}{i} = {value}')
                elif '.' in i:
                    # class attribute access
                    spaces += len(outer)
                    instance = i.split('.')[0]
                    obj = local_var.get(instance, global_var.get(instance))
                    attribute = get_recur_attr(obj, i[len(instance) + 1:])
                    value = get_repr(attribute, spaces)
                    scope = outer if instance in global_var else ''
                    txt.append(f'     {bullet_1} {scope}{i} = {value}')
                else:
                    # reserved Keyword or non-identifier, eg. word inside the string
                    pass

            # get value of other variables within local scope
            if full_context:
                other_local_var = set(local_var) - set(identifiers)
                if other_local_var:
                    spaces = multi_line_indent2
                    txt.extend([f'     {bullet_2} {k} = {get_repr(v, spaces + len(k))}'
                                for k, v in local_var.items() if k in other_local_var])

            txt.append('')
            result.append('\n'.join(txt))

    return '\n'.join(result)



import time
import sys
from random import randint

from multiprocessing import Pool
from logger_tt import setup_logging
from logging import getLogger


__author__ = "Duc Tin"
logger = getLogger(__name__)
setup_logging(use_multiprocessing=True)


def worker(arg):
    logger.info(f'child process {arg}: started')
    time.sleep(randint(1,10)/10)
    logger.info(f'child process {arg}: stopped')


if __name__ == '__main__':
    if len(sys.argv) > 1:
        proc_no = int(sys.argv[1])
    else:
        proc_no = 7

    logger.info('Parent process pool is ready to spawn child')

    with Pool(4) as p:
        p.map(worker, range(proc_no))

    print('__finished__')



import sys
import logging
import json
import threading
from pathlib import Path
from logging.config import dictConfig
from logging import getLogger
from .inspector import analyze_frame, logging_disabled
from .core import LogConfig, DefaultFormatter
from multiprocessing import current_process

__author__ = "Duc Tin"
__all__ = ['setup_logging', 'logging_disabled', 'getLogger', 'logger']

"""Config log from file and make it also logs uncaught exception"""

internal_config = LogConfig()


def handle_exception(exc_type, exc_value, exc_traceback, thread_name=''):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    full_context = internal_config.full_context

    # Root logger will log all other uncaught exceptions
    txt = analyze_frame(exc_traceback, full_context)

    # Exception in a child thread?
    if thread_name:
        thread_name = ' in ' + thread_name

    logging.error(f"Uncaught exception{thread_name}:\n"
                  f"Traceback (most recent call last):\n"
                  f"{txt}",
                  exc_info=(exc_type, exc_value, None))

    if not thread_name:
        # As interpreter is going to shutdown after this function,
        # objects are getting deleted.
        # Disable further logging to prevent NameError exception
        logging.disable(logging.CRITICAL)
        # Don't do this if exception is on child thread


def thread_run_with_exception_logging(self):
    """Do everything the Thread.run() do and add exception handling"""
    try:
        if self._target:
            self._target(*self._args, **self._kwargs)
    except Exception:
        handle_exception(*sys.exc_info(), thread_name=self.name)
    finally:
        # Avoid a refcycle if the thread is running a function with
        # an argument that has a member that points to the thread.
        del self._target, self._args, self._kwargs


def ensure_path(config: dict, override_log_path: str = ""):
    """ensure log path exists"""
    for handler in config['handlers'].values():
        filename = handler.get('filename')
        if not filename:
            continue
        filename = override_log_path or filename
        handler['filename'] = filename
        log_path = Path(filename).parent
        log_path.mkdir(parents=True, exist_ok=True)


def load_from_file(f: Path) -> dict:
    if f.suffix in ['.yaml', '.yml']:
        try:
            import yaml  # will raise error if pyyaml is not installed
            safe_load = yaml.safe_load
        except ImportError:
            try:
                from ruamel.yaml import YAML
                yaml = YAML(typ='safe')
                safe_load = yaml.load
            except ImportError:
                raise ImportError('Required package not found: "pyyaml" or "ruamel.yaml"')

        dict_cfg = safe_load(f.read_text())

    else:
        with f.open() as fp:
            dict_cfg = json.load(fp)

    # add default formatters to use logger_tt logger right on spot
    try:
        dlf = dict_cfg['logger_tt'].pop('default_logger_formats', {})
    except KeyError:
        dlf = {}
    DefaultFormatter.default_formats.update(dlf)
    for formatter in dict_cfg['formatters'].values():
        if not formatter.get('class'):
            formatter['class'] = 'logger_tt.core.DefaultFormatter'

    return dict_cfg


def merge_config(from_file: dict, from_func: dict) -> dict:
    """Override logger_tt config of from_file by
        the argument passed to the setup_logging function
    """
    defaults = dict(capture_print=False, strict=False, guess_level=False,
                    full_context=0, suppress=None,
                    suppress_level_below=logging.WARNING, use_multiprocessing=False)
    merged = {}
    for key, val in defaults.items():
        merged[key] = from_func.get(key, from_file.get(key, val))

    # check for unknown key
    uff1 = set(from_file) - set(defaults)
    uff2 = set(from_func) - set(defaults)

    if uff1:
        raise TypeError(f'setup_logging() got an unexpected keyword argument(s): {uff1}')
    if uff2:
        raise ValueError(f'logger_tt unknown fields: {uff2}')

    return merged


def setup_logging(config_path: str = "", log_path: str = "", **logger_tt_config) -> LogConfig:
    """Setup logging configuration

    :param config_path: Path to log config file. Use default config if this is not provided
    :param log_path: Path to store log file. Override 'filename' field of 'handlers' in
        default config.
    :param logger_tt_config: keyword only arguments to config the logger. Fields in this dictionary
        will override the same field in the config file.
        :key capture_print: bool, log message that is printed out with print() function
        :key strict       : bool, only used when capture_print is True.
                            If strict is True, then log everything that use sys.stdout.write().
        :key guess_level  : bool, auto guess logging level of captured message
        :key full_context : int, whether to log full local scope on exception or not and up to what level
        :key suppress     : list[str], name of loggers to be suppressed.
        :key suppress_level_below: int, for logger in the suppress list,
                                    any message below this level is not processed, not printed out nor logged to file
        :key use_multiprocessing : bool or str, set this to True if your code use multiprocessing.
                                    This flag switches the queue used for logging from
                                    queue.Queue to multiprocessing.Queue . This option can only be used here.
    """
    if internal_config.initialized:
        logger.warning('Re-initializing logger_tt. "setup_logging()" should only be called one.')

    if config_path:
        cfgpath = Path(config_path)
        assert cfgpath.is_file(), 'Input config path is not a file!'
        assert cfgpath.suffix in ['.yaml', '.json', '.yml'], 'Config file type must be either yaml, yml or json!'
        assert cfgpath.exists(), f'Config file path not exists! {cfgpath.absolute()}'
    else:
        cfgpath = Path(__file__).parent / 'log_config.json'

    # load config from file
    config = load_from_file(cfgpath)
    ensure_path(config, log_path)
    logger_tt_cfg = config.pop('logger_tt', {})

    # initialize
    for name in logging.root.manager.loggerDict:
        existing_logger = logging.getLogger(name)
        existing_logger.__class__ = ExceptionLogger
    else:
        logging.config.dictConfig(config)

    if current_process().name == 'MainProcess':
        logging.debug('New log started'.center(50, '_'))
        logging.debug(f'Log config file: {cfgpath}')

    # set internal config
    iconfig = merge_config(logger_tt_cfg, logger_tt_config)
    internal_config.from_dict(iconfig)

    # capture other messages
    sys.excepthook = handle_exception
    threading.Thread.run = thread_run_with_exception_logging
    return internal_config


class ExceptionLogger(logging.Logger):
    """Modify the `exception` func so that it print out context too
        This allow user do a try-except in outer code but still has the full log
        of nested code's error
        Example:
            try:
                a, b = 1, 0
            except Exception as e:
                logger.exception(e)
            # then move on
    """

    _logger_names = {}

    def exception(self, msg, *args, exc_info=True, **kwargs):
        if exc_info:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            full_context = internal_config.full_context
            txt = analyze_frame(exc_traceback, full_context)
            logging.error(f'{msg}\n'
                          f"Traceback (most recent call last):\n"
                          f"{txt}",
                          exc_info=(exc_type, exc_value, None))
        else:
            logging.error(msg, *args, exc_info=exc_info, **kwargs)

    def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
        record = super().makeRecord(name, level, fn, lno, msg, args, exc_info, func, extra, sinfo)

        if name == 'logger_tt':
            # try to get the __name__ of the module that use the default logger: logger_tt
            pathname = fn.replace('\\', '/')
            qualified_name = self._logger_names.get(pathname)
            if not qualified_name:
                for qualified_name, module in sys.modules.items():
                    file = getattr(module, '__file__', None)
                    if file and file.replace('\\', '/') == pathname:
                        self._logger_names[pathname] = qualified_name
                        break

            if qualified_name == '__main__' and record.processName != 'MainProcess':
                qualified_name = '__mp_main__'

            record.filename = qualified_name or record.filename

        return record


def logger_tt_filter(record):
    if record.filename not in internal_config.suppress_list:
        return True
    if record.levelno > internal_config.suppress_level_below:
        return True


logging.setLoggerClass(ExceptionLogger)
logger = getLogger('logger_tt')  # pre-made default logger for all modules
logger.setLevel(logging.DEBUG)
logger.addFilter(logger_tt_filter)



from logger_tt import setup_logging
from logging import getLogger


__author__ = "Duc Tin"
setup_logging(full_context=1)
logger = getLogger(__name__)


class Dummy:
    def __init__(self):
        self.value = 3
        self.divisor = 0

    def __str__(self):
        return 'Dummy(my dummy class)'


def my_function(var_in, *arg, **kwargs):
    my_local_var = 345
    raise RuntimeError


if __name__ == '__main__':
    a = Dummy()
    my_function(a, 456, 789, my_kw='hello', another_kw='world')



from tests.subsub_module import error_func
from logger_tt import setup_logging, logger, getLogger

__author__ = "Duc Tin"

normal_logger = getLogger(__name__)


def run():
    logger.info('Logging by a default logger')
    normal_logger.info('Logging by the normal logger')


if __name__ == '__main__':
    setup_logging()
    run()
    error_func()



from setuptools import setup

with open('README.md', encoding='utf8') as fi:
    long_description = fi.read()

setup(
    name='logger_tt',
    version='1.6.0',
    packages=['logger_tt'],
    url='https://github.com/Dragon2fly/logger_tt',
    package_data={'': ['log_config.json', 'log_config.yaml']},
    include_package_data=True,
    license='MIT',
    platforms=["Any platform -- don't need Windows"],
    author='Nguyen Ba Duc Tin',
    author_email='nguyenbaduc.tin@gmail.com',
    description='Make logging simple, log even exception that you forgot to catch. '
                'Make multiprocessing logging becomes a breeze',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: OS Independent',
        'Topic :: System :: Hardware :: Symmetric Multi-processing',
        'Topic :: System :: Logging',
        'Topic :: Software Development :: Debuggers',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires=">=3.6",
)
